/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University. 
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2018-07-31
 */

#include <math.h>
#include "T2TAttention.h"
#include "T2TUtility.h"
#include "T2TEmbedding.h"
#include "../../tensor/core/CHeader.h"

namespace transformer
{

enum { NONE, SELF, CONTEXT };

/* constructor */
T2TAttention::T2TAttention()
{
    nhead = -1;
    dk = -1;
    dv = -1;
    d  = -1;
    isMasked = false;
    ignored = 0;
}

/* deconstructor */
T2TAttention::~T2TAttention()
{
}

/* 
initialize the model 
>> argc - number of arguments
>> argv - list of pointers to the arguments
>> myIgnored - number of position ignored in attention (from the begining)
>> myIsMasked - indicates whether the attention is with a mask
>> myDevID - device id
>> myMem - the memory pool
*/
void T2TAttention::InitModel(int argc, char ** argv, 
                             bool myIsMasked, int myIgnored, 
                             int myDevID, XMem * myMem)
{
    devID = myDevID;
    mem = myMem;
    isMasked = myIsMasked;
    ignored = myIgnored;
    
    float minmax = 0;

    LoadParamInt(argc, argv, "nhead", &nhead, 8);
    LoadParamInt(argc, argv, "d", &dk, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "d", &dv, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "d", &d, DEFAULT_EMBEDDING_SIZE);
    LoadParamInt(argc, argv, "maxposition", &maxRP, 8);
    LoadParamFloat(argc, argv, "attminmax", &minmax, 0.1F);
    LoadParamFloat(argc, argv, "dropoutatt", &dropoutP, 0);

    InitTensor2DV2(&wq, d, d, X_FLOAT, devID);
    InitTensor1DV2(&bq, d, X_FLOAT, devID);
    InitTensor2DV2(&wk, d, d, X_FLOAT, devID);
    InitTensor1DV2(&bk, d, X_FLOAT, devID);
    InitTensor2DV2(&wv, d, d, X_FLOAT, devID);
    InitTensor1DV2(&bv, d, X_FLOAT, devID);
    InitTensor2DV2(&rpEmbK, maxRP * 2 + 1, d/nhead, X_FLOAT, devID);
    InitTensor2DV2(&wa, d, d, X_FLOAT, devID);
    InitTensor1DV2(&ba, d, X_FLOAT, devID);
}

/* 
make the network 
>> k - keys. It might be of size B * L * H
       where B = batch size, L = sequence length, 
       and H = vector size of each position
>> q - queries
>> v - values
>> mask - as it is
>> isTraining - indicates whether the model is used for training
>> cache - layer cache list
>> cacheType - which type that cache is
<< return - multi-attention result
*/
XTensor T2TAttention::Make(XTensor& k, XTensor& q, XTensor& v, XTensor *mask, 
                           bool isTraining, Cache * cache, int cacheType)
{
    bool isEnc = (!cache) ? true : false;
    int k2Dim[]{ k.GetDim(0),k.GetDim(1), wk.GetDim(1) };
    int v2Dim[]{ v.GetDim(0),v.GetDim(1), wv.GetDim(1) };
    XTensor* q2 = NewTensor3DV2(q.GetDim(0),q.GetDim(1), wq.GetDim(1), X_FLOAT, q.devID);
    XTensor* k2 = NULL;
    XTensor* v2 = NULL;
    XTensor* kNewCache = NULL;
    XTensor* vNewCache = NULL;

    /* linear transformation before self-attention */
    /* notice that all weights are transposed!!! */

    _MatrixMul(&q, X_NOTRANS, &wq, X_TRANS, q2);
    _SumDim(q2, &bq, 2);

    if (!cache) {
        k2 = NewTensor3DV2(k.GetDim(0),k.GetDim(1), wk.GetDim(1), X_FLOAT, k.devID);
        v2 = NewTensor3DV2(v.GetDim(0),v.GetDim(1), wv.GetDim(1), X_FLOAT, v.devID);
        _MatrixMul(&k, X_NOTRANS, &wk, X_TRANS, k2);
        _SumDim(k2, &bk, 2);
        _MatrixMul(&v, X_NOTRANS, &wv, X_TRANS, v2);
        _SumDim(v2, &bv, 2);
    }
    else {
        if (cacheType == SELF) {
            k2 = NewTensor3DV2(q.GetDim(0), q.GetDim(1), wk.GetDim(1), X_FLOAT, q.devID);
            v2 = NewTensor3DV2(q.GetDim(0), q.GetDim(1), wv.GetDim(1), X_FLOAT, q.devID);
            _MatrixMul(&q, X_NOTRANS, &wk, X_TRANS, k2);
            _SumDim(k2, &bk, 2);
            _MatrixMul(&q, X_NOTRANS, &wv, X_TRANS, v2);
            _SumDim(v2, &bv, 2);
            if (!cache->IsEmpty()) {
                XTensor* kOldCache = cache->GetK();
                XTensor* vOldCache = cache->GetV();
                kNewCache = NewTensor3DV2(kOldCache->GetDim(0), kOldCache->GetDim(1) + k2->GetDim(1), kOldCache->GetDim(2), X_FLOAT, k2->devID);
                vNewCache = NewTensor3DV2(vOldCache->GetDim(0), vOldCache->GetDim(1) + v2->GetDim(1), vOldCache->GetDim(2), X_FLOAT, v2->devID);
                _Concatenate(kOldCache, k2, kNewCache, 1);
                _Concatenate(vOldCache, v2, vNewCache, 1);
                DelTensor(k2);
                DelTensor(v2);
                k2 = kNewCache;
                v2 = vNewCache;
            }
            cache->Update(k2, v2);
        }
        else if (cacheType == CONTEXT) {
            if (cache->IsEmpty()) {
                k2 = NewTensor3DV2(k.GetDim(0), k.GetDim(1), wk.GetDim(1), X_FLOAT, k.devID);
                v2 = NewTensor3DV2(v.GetDim(0), v.GetDim(1), wv.GetDim(1), X_FLOAT, v.devID);
                _MatrixMul(&k, X_NOTRANS, &wk, X_TRANS, k2);
                _SumDim(k2, &bk, 2);
                _MatrixMul(&v, X_NOTRANS, &wv, X_TRANS, v2);
                _SumDim(v2, &bv, 2);
                cache->Update(k2, v2);
            }
            else {
                k2 = cache->GetK();
                v2 = cache->GetV();
            }
        }
        else {
            CheckNTErrors(0, "invalid cache type");
        }
    }

    if(cacheType == CONTEXT)
        return MakeAttention(k2, q2, v2, mask, isTraining, isEnc);

    return MakeRPRAttention(k2, q2, v2, mask, isTraining, isEnc);
}
    
/*
make the attention network given keys, queries and values (after linear transformation)
>> k - keys. It might be of size B * L * H
       where B = batch size, L = sequence length,
       and H = vector size of each position
>> q - queries
>> v - values
>> mask - as it is
>> isTraining - indicates whether the model is used for training
*/
XTensor T2TAttention::MakeAttention(XTensor *k, XTensor *q, XTensor *v, const XTensor *mask, bool isTraining, bool isEnc)
{
    XTensor kheads;
    XTensor qheads;
    XTensor vheads;
    
    /* multi head */
    kheads = Split(*k, k->order - 1, nhead);
    qheads = Split(*q, q->order - 1, nhead);
    vheads = Split(*v, v->order - 1, nhead);
    
    XTensor att; 
    XTensor dot;
    XTensor scalar;
    
    /* scalar = softmax(Q * K^T / sqrt(dk)) * V */
    dot = BMMul(qheads, X_NOTRANS, kheads, X_TRANS);
    
    if (isMasked && mask) {
        _SumMe(&dot, mask);
    }
    
    dot = Linear(dot, 1.0F / (float)sqrt((float)dk / nhead));
    
    scalar = Softmax(dot, -1);

    if(isTraining && dropoutP > 0)
        scalar = Dropout(scalar, dropoutP);

    att = BMMul(scalar, vheads);

    /* concatenate the heads */
    return MulAndShift(Merge(att, att.order - 1), X_NOTRANS, wa, X_TRANS, ba);
}

/*
make the attention network by incorporating the relative position representation with the given keys, queries and values (after linear transformation)
>> k - keys. It might be of size B * L * H
where B = batch size, L = sequence length,
and H = vector size of each position
>> q - queries
>> v - values
>> mask - as it is
>> isTraining - indicates whether the model is used for training
*/
XTensor T2TAttention::MakeRPRAttention(XTensor *k, XTensor *q, XTensor *v, XTensor *mask, bool isTraining, bool isEnc)
{
    XTensor kheads;
    XTensor qheads;
    XTensor vheads;

    const int batchSize = q->GetDim(0);
    const int lenQ = q->GetDim(1);
    const int lenKV = k->GetDim(1);

    /* multi head */
    kheads = Split(*k, k->order - 1, nhead);
    qheads = Split(*q, q->order - 1, nhead);
    vheads = Split(*v, v->order - 1, nhead);

    XTensor att;
    XTensor dot;
    XTensor scalar;

    XTensor embMatrix, relativeKey;
    InitTensor2DV2(&embMatrix, lenQ, lenKV, X_INT, q->devID);
    InitTensor3DV2(&relativeKey, lenQ, lenKV, kheads.GetDim(-1), X_FLOAT, q->devID);
    InitTensor4DV2(&dot, nhead, batchSize, lenQ, lenKV, X_FLOAT, q->devID);

    /* generate the relative emb index (L_q, L_kv) */
    GetRPEmbedding(&embMatrix, lenQ, lenKV, maxRP, q->devID,isEnc);

    
    /* generate the relative key from the rpEmbK (L_q, L_kv, H/K) */
    _Gather(&rpEmbK, &relativeKey, &embMatrix);

    /* RPR dot product (K, B, L_q, L_kv)*/
    qheads = qheads / float(nhead);
    
    RPDotProduct(&qheads, &kheads, &relativeKey, &dot, true);

    if (isMasked && mask)
        _SumMe(&dot, mask);

    /* scale the dot result */
    //dot = Linear(dot, 1.0F / (float)sqrt((float)dk / nhead));

    /* softmax */
    scalar = Softmax(dot, -1);

    if (isTraining && dropoutP > 0)
        scalar = Dropout(scalar, dropoutP);

    /* generate the relative attention output (K, B, L_q, H/K) */
    att = BMMul(scalar, vheads);

    if (isEnc) {
        DelTensor(k);
        DelTensor(q);
        DelTensor(v);
    }

    /* concatenate the heads */
    return MulAndShift(Merge(att, att.order - 1), X_NOTRANS, wa, X_TRANS, ba);
}

void T2TAttention::GetRPEmbedding(XTensor* embMatrix, const int lenQ, const int lenKV, const int maxRelativeLen, const int devID, const bool isEnc)
{

    XTensor range;
    InitTensor1DV2(&range, lenKV, X_INT, devID);
    int* index = new int[lenKV];
    // for encoder self-attention which the L_q = L_kv
    if (isEnc)
    {
        for (int i = 0; i <lenKV; i++)
            index[i] = i;
        range.SetData(index, lenKV);
        XTensor range_2D, range_2D_t;
        InitTensor2DV2(&range_2D, lenQ, lenKV, X_INT, devID);
        InitTensor2DV2(&range_2D_t, lenQ, lenKV, X_INT, devID);
        _Unsqueeze(&range, &range_2D, 0, lenQ);
        _Transpose(&range_2D, &range_2D_t, 0, 1);
        _Sum(&range_2D, &range_2D_t, embMatrix, -1);
    }
    // for decoder self-attention which the L_q != L_kv, and L_q is 1
    else
    {
        for (int i = 0; i <lenKV; i++)
            index[i] = -lenKV + i + 1;
        range.SetData(index, lenKV);
        _Unsqueeze(&range, embMatrix, 0, lenQ);
    }

    // clip the tensor range from -max_ralative_length to maxRelativeLen
    _Clip(embMatrix, embMatrix, -maxRelativeLen, maxRelativeLen);

    // (L_q, L_kv)
    _ScaleAndShift(embMatrix, embMatrix, 1, maxRelativeLen);
    delete[] index;
}

void T2TAttention::RPDotProduct(XTensor* x, XTensor* y, XTensor* z, XTensor* attention, const bool isKey)
{

    const int heads = nhead;
    const int batchSize = x->dimSize[1];
    const int lenKV = y->dimSize[2];
    const int lenQ = x->dimSize[2];
    const int depth = y->dimSize[3];
    // L_kv (isKey=True) or H/K (isKey=False)
    const int last_dim = isKey ? lenKV : depth;
    MATRIX_TRANS_TYPE transpose_flag = isKey ? X_TRANS : X_NOTRANS;

    //if (profiler_) profiler_->StartTimer("RPDotPro-BMM");
    // for key: batch-MM: (K,B,L_q,H/K) * (K,B,H/K,L_kv) -> (K,B,L_q,L_kv)
    // for not key: batch-MM: (K,B,L_q,L_kv) * (K,B,L_kv,H/K) -> (K,B,L_q,H/K)
    XTensor context;
    InitTensor4DV2(&context, heads, batchSize, lenQ, last_dim, X_FLOAT, x->devID);
    _MatrixMulBatched(x, X_NOTRANS, y, transpose_flag, &context);
    //if (profiler_) profiler_->FinishTimer("RPDotPro-BMM");

    // reshape and transpose x to (L_q, K*B, H/K or L_kv)
    int merge_dims[] = { heads * batchSize, lenQ, x->dimSize[3] };
    x->Reshape(3, merge_dims);
    XTensor xTrans;
    InitTensor3DV2(&xTrans, lenQ, heads * batchSize, x->GetDim(-1), X_FLOAT, x->devID);
    _Transpose(x, &xTrans, 0, 1);

    // for key: batch-MM: (L_q, K*B, H/K) * (L_q, L_kv, H/K) -> (L_q, K*B, L_kv)
    // for not key: batch-MM: (L_q, K*B, L_kv) * (L_q, L_kv, H/K) -> (L_q, K*B, H/K)
    XTensor relative;
    InitTensor3DV2(&relative, lenQ, heads * batchSize, last_dim, X_FLOAT, x->devID);
    _MatrixMulBatched(&xTrans, X_NOTRANS, z, transpose_flag, &relative);

    // (L_q, K*B, H/K or L_kv) -> (K*B, L_q, H/K or L_kv)
    XTensor relative_t;
    InitTensor3DV2(&relative_t, heads * batchSize, lenQ, last_dim, X_FLOAT, x->devID);
    _Transpose(&relative, &relative_t, 0, 1);

    int split_dims[] = { heads, batchSize, lenQ, last_dim };
    relative_t.Reshape(4, split_dims);

    _Sum(&context, &relative_t, attention);
}
}
