/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2019, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

 /*
  * $Created by: XIAO Tong (xiaotong@mail.neu.edu.cn) 2019-03-27
  */

#include <math.h>

#include "T2TTester.h"
#include "T2TSearch.h"
#include "T2TUtility.h"
#include "../../tensor/XUtility.h"
#include "../../tensor/core/CHeader.h"

using namespace nts;

namespace transformer
{

/* constructor */
T2TTester::T2TTester()
{

}

/* de-constructor */
T2TTester::~T2TTester()
{
}

/* initialize the model */
void T2TTester::Init(int argc, char** argv)
{
    LoadParamInt(argc, argv, "vsize", &vSize, 1);
    LoadParamInt(argc, argv, "vsizetgt", &vSizeTgt, vSize);
    LoadParamInt(argc, argv, "sentbatch", &sentBatch, 1);
    LoadParamBool(argc, argv, "sort", &batchLoader.sortBuffer, false);

    seacher.Init(argc, argv);
}

Result ExtractRes(XTensor& output, IntList& indices, int i) {
    Result res;
    XTensor sent, srcIdx, tgtIdx;
    InitTensor1D(&srcIdx, 1, X_INT, output.devID);
    int idx[]{ i };
    srcIdx.SetData(idx, 1);
    InitTensor(&tgtIdx, &srcIdx);
    tgtIdx.SetAscendingOrder(0);

    sent = CopyIndexed(output, 0, srcIdx, tgtIdx);

    res.data.Add((int*)sent.data, sent.unitNum);
    res.id = indices[i];
    return res;
}

/*
test the model
>> fn - test data file
>> ofn - output data file
>> model - model that is trained
*/
void T2TTester::Test(const char* fn, const char* ofn, T2TModel* model)
{
    int wc = 0;
    int wordCount = 0;
    int wordCountTotal = 0;
    int sentCount = 0;
    int batchCount = 0;

    /* data files */
    FILE* ofile = fopen(ofn, "w");
    CheckNTErrors(ofile, "Cannot open the output file");

    int devID = model->devID;
    XMem* mem = model->mem;

    double startT = GetClockSec();

    /* batch of input sequences */
    XTensor batchEnc;

    /* padding */
    XTensor paddingEnc;

    /* an array that keeps the sequences */
    int* seqs = new int[MILLION];

    batchLoader.Init(fn, 100, true);

    int count = 0;
    while (!batchLoader.IsEmpty()) {
        wordCount = 0;

        /* reset cache for decoder */
        model->ResetCache();
        
        IntList indices = batchLoader.LoadBatch(&batchEnc, &paddingEnc, sentBatch, devID);

        XTensor output, score;

        seacher.Search(model, &batchEnc, &paddingEnc, &output, &score);

        for (int i = 0; i < indices.Size(); i++)
            batchLoader.resBuffer.Add(ExtractRes(output, indices, i));

        wc = batchEnc.GetDim(-1);
        wordCount += wc;
        wordCountTotal += wc;
        sentCount += batchEnc.GetDim(-2);
        batchCount += 1;

        double elapsed = GetClockSec() - startT;
        XPRINT3(0, stderr, "[INFO] elapsed=%.1fs, sent=%d, sword=%d\n", elapsed, sentCount, wordCount);
    }

    batchLoader.SortRes();

    for (int i = 0; i < batchLoader.resBuffer.Size(); i++)
        Dump(ofile, batchLoader.resBuffer[i].data);

    fclose(ofile);
    delete[] seqs;

    double elapsed = GetClockSec() - startT;
    XPRINT3(0, stderr, "[INFO] test finished (took %.1fs, word=%d, sent=%d)\n", elapsed, wordCountTotal, sentCount);
}

/*
dump the result into the file
>> file - data file
>> output - output list
*/
void T2TTester::Dump(FILE* file, IntList& output)
{
    for (int i = 0; i < output.Size(); i++) {
        int w = output[i];
        if (w < 0) {
            if (i == 0)
                return;
            else
                break;
        }
            
        fprintf(file, "%d ", w);
    }
    fprintf(file, "\n");
}

}
