/* NiuTrans.Tensor - an open-source tensor library
 * Copyright (C) 2018, Natural Language Processing Lab, Northestern University.
 * All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
* $Created by: HU Chi (huchinlp@foxmail.com) 2019-04-05
*/

#include <string>
#include <fstream>
#include <algorithm>
#include "DataSet.h"
#include "StringUtil.h"
#include "../../../tensor/XUtility.h"

using namespace nts;
using namespace std;

/* sort results by their ids */
void DataSet::SortRes()
{
    auto cmp = [](Result& a, Result& b) {
        return a.id < b.id; 
    };
    std::sort(resBuffer.items, resBuffer.items + resBuffer.count, cmp);
}

/*
load data from the file to the buffer
*/
void DataSet::LoadDataToBuffer()
{
    bufferUsed = 0;
    srcBuffer.Clear();
    
    bufferSize = min(bufferSize, exampleNumber);

    for (int i = 0; i < bufferSize; i++) {
        
        long off = offset[index++];
        IntList data(off);
        data.count = off;
        fread(data.items, sizeof(int), off, fp);
        
        Example example;
        example.id = id++;
        example.data = data;
        srcBuffer.Add(example);
    }

    if (sortBuffer) {
        auto cmp = [](Example& a, Example& b) {
            return a.data.Size() > b.data.Size(); 
        };
        std::sort(srcBuffer.items, srcBuffer.items + srcBuffer.count, cmp);
    }
}

/*
select a field and generate a mini-batch by indices
>>> batchEnc - a tensor to store the batch of input
>>> paddingEnc - a tensor to store the batch of paddings
>>> batchSize - batch size
>>> devID - devices id, -1 for CPU
>>> mem - the memory pool
*/
IntList DataSet::LoadBatch(XTensor * batchEnc, XTensor * paddingEnc, size_t batchSize, int devID)
{
    if(srcBuffer.count == 0)
        LoadDataToBuffer();

    size_t realBatchSize = batchSize;

    /* real batch size */
    if ((srcBuffer.Size() - bufferUsed) < batchSize) {
        realBatchSize = srcBuffer.Size() - bufferUsed;
    }

    /* get the maximum sentence length in a mini-batch */
    size_t maxLen = 0;
    if (realBatchSize == 1) {
        maxLen = srcBuffer[bufferUsed].data.Size();
    }

    for (size_t i = 0; i < realBatchSize - 1; i++) {
        maxLen = max(maxLen, srcBuffer[bufferUsed + i].data.Size());
    }
    CheckNTErrors(maxLen != 0, "wrong length dectected");

    int* batchValues = new int[maxLen * realBatchSize];
    float* paddingValues = new float[maxLen * realBatchSize];

    for (int i = 0; i < realBatchSize * maxLen; i++) {
        batchValues[i] = 1.0F;
    }
    memset(batchValues, 0, sizeof(int) * maxLen * realBatchSize);
    memset(paddingValues, 0, sizeof(float) * maxLen * realBatchSize);

    size_t cur = 0;
    
    /* left padding */
    IntList indices;
    indices.Reserve(realBatchSize);

    for (size_t i = 0; i < realBatchSize; i++) {

        indices.Add(srcBuffer[bufferUsed + i].id);
        IntList& data = srcBuffer[bufferUsed + i].data;

        cur = maxLen * (i + 1) - data.Size();
        for (int j = 0; j < data.Size(); j++) {
            batchValues[cur] = data[j];
            paddingValues[cur++] = 1.0F;
        }
        cur = maxLen * (i + 1);
    }

    InitTensor2DV2(batchEnc, realBatchSize, maxLen, X_INT, devID);
    InitTensor2DV2(paddingEnc, realBatchSize, maxLen, X_FLOAT, devID);

    bufferUsed += realBatchSize;

    batchEnc->SetData(batchValues, batchEnc->unitNum);
    paddingEnc->SetData(paddingValues, paddingEnc->unitNum);
    delete[] batchValues;
    delete[] paddingValues;

    return indices;
}

/* 
convert text file to binary file
format of the text file: 
one sentence per line, seperated by a blank
format of the binary file: 
part 1: number of all examples
part 2: offsets of all examples
part 3: the raw data 
>>> src - the path of source text file
>>> tgt - the path of target binary file
*/
void nts::DataSet::ConvertFile(const char* src, const char* tgt)
{
    ifstream ifile(src, ios::in);
    FILE* ofile = fopen(tgt, "wb");

    CheckNTErrors(ofile, "unable to create the output file");

    string line;
    size_t idx = 0;
    const int maxExample = 10240;
    IntList dataList[maxExample];
    
    while (getline(ifile, line)){
        SplitInt(line, " ", dataList[idx++]);
    }
    
    /* part 1: number of examples */
    fwrite(&idx, sizeof(idx), 1, ofile);

    /* part 2: offset of all examples */
    for (int i = 0; i < idx; i++) {
        size_t size = (dataList[i].Size());
        fwrite(&size, sizeof(size), 1, ofile);
    }

    /* part 3: value of examples */
    for (int i = 0; i < idx; i++) {
        fwrite(dataList[i].items, sizeof(int), dataList[i].Size(), ofile);
    }

    ifile.close();
    fclose(ofile);
}

/*
the constructor of DataSet
the binary data consists of three parts
part 1: number of all examples
part 2: offsets of all examples
part 3: the raw data 
>>> fname - path of the data file
>>> myBufferSize - size of the data buffer
>>> mySortBuffer - whether sort the data
*/
void DataSet::Init(const char* fname, size_t myBufferSize, bool mySortBuffer)
{
    id = 0; 
    index = 0;
    bufferUsed = 0;
    bufferSize = myBufferSize;
    sortBuffer = mySortBuffer;
    
    fp = fopen(fname, "rb");
    
    CheckNTErrors(fp, "can not open the file");

    /* read offsets */
    exampleNumber = 0;
    fread(&exampleNumber, sizeof(exampleNumber), 1, fp);
    CheckNTErrors(exampleNumber > 0, "invalid example numbers");
    offset.Reserve(exampleNumber);
    for (int i = 0; i < exampleNumber; i++) {
        size_t off;
        fread(&off, sizeof(off), 1, fp);
        offset.Add(off);
    }

    /* reset the buffer size if it is too big */
    bufferSize = min(bufferSize, exampleNumber);

    srcBuffer.Reserve(bufferSize);
}

/* check if the buffer is empty */
bool nts::DataSet::IsEmpty()
{
    return (index >= offset.count) && (bufferUsed >= bufferSize);
}

/* de-constructor */
nts::DataSet::~DataSet()
{
    if (fp) {
        fclose(fp);
    }
}